@@ -2,17 +2,23 @@ package main
 
 import (
 	"bufio"
+	"context"
 	"fmt"
-	"github.com/jackc/pgx"
 	"os"
-	"time"
+
+	"github.com/jackc/pgx"
 )
 
 var pool *pgx.ConnPool
 
 func main() {
-	var err error
-	pool, err = pgx.NewConnPool(extractConfig())
+	config, err := pgx.ParseEnvLibpq()
+	if err != nil {
+		fmt.Fprintln(os.Stderr, "Unable to parse environment:", err)
+		os.Exit(1)
+	}
+
+	pool, err = pgx.NewConnPool(pgx.ConnPoolConfig{ConnConfig: config})
 	if err != nil {
 		fmt.Fprintln(os.Stderr, "Unable to connect to database:", err)
 		os.Exit(1)
@@ -58,38 +64,12 @@ func listen() {
 	conn.Listen("chat")
 
 	for {
-		notification, err := conn.WaitForNotification(time.Second)
-		if err == pgx.ErrNotificationTimeout {
-			continue
-		}
+		notification, err := conn.WaitForNotification(context.Background())
 		if err != nil {
 			fmt.Fprintln(os.Stderr, "Error waiting for notification:", err)
 			os.Exit(1)
 		}
 
-		fmt.Println("PID:", notification.Pid, "Channel:", notification.Channel, "Payload:", notification.Payload)
+		fmt.Println("PID:", notification.PID, "Channel:", notification.Channel, "Payload:", notification.Payload)
 	}
 }
-
-func extractConfig() pgx.ConnPoolConfig {
-	var config pgx.ConnPoolConfig
-
-	config.Host = os.Getenv("CHAT_DB_HOST")
-	if config.Host == "" {
-		config.Host = "localhost"
-	}
-
-	config.User = os.Getenv("CHAT_DB_USER")
-	if config.User == "" {
-		config.User = os.Getenv("USER")
-	}
-
-	config.Password = os.Getenv("CHAT_DB_PASSWORD")
-
-	config.Database = os.Getenv("CHAT_DB_DATABASE")
-	if config.Database == "" {
-		config.Database = "postgres"
-	}
-
-	return config
-}
